/*
  Copyright (c) 2014, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "Map.hpp"

#include <cstdio>
#include <cstdlib>
#include <fstream>


namespace RPGEdit {

namespace Model {

Position<float> Map::getEntityPosition(const Entity *entity, uint64_t time) const {
	const EntityState &state = entityStates.at(entity);

	Position<float> p(state.position);

	if (state.transitionEnd)
		if (time <= state.transitionStart)
			return p;
		else if (time >= state.transitionEnd)
			return p + state.direction;
		else
			return p.translate(state.direction,
					   float(time-state.transitionStart)/
					   float(state.transitionEnd-state.transitionStart));
	else
		return p;
}

bool Map::moveEntity(Entity *entity, Direction dir, uint64_t start, uint64_t end) {
	EntityState &state = entityStates.at(entity);

	if (state.transitionEnd)
		return false;

	entity->setDirection(dir);

	if (isBlocked(state.position + dir))
		return false;

	state.transitionStart = start;
	state.transitionEnd = end;
	state.direction = dir;

	return true;
}

void Map::moveEntityTo(Entity *entity, Position<int> pos) {
	EntityState &state = entityStates.at(entity);

	removeEntityPosition(state.position, entity);
	addEntityPosition(pos, entity);

	state.position = pos;
	state.transitionStart = state.transitionEnd = 0;
}

void Map::finishEntityTransition(Entity *entity) {
	EntityState &state = entityStates.at(entity);

	if (state.transitionEnd)
		moveEntityTo(entity, state.position + state.direction);
}

std::unique_ptr<Map> Map::load(__attribute__((unused)) const std::string &name) {
	std::string filename = "../resources/map/" + name + ".map";

	std::ifstream file;

	file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
	file.open(filename.c_str());

	std::string line;

	std::getline(file, line);
	size_t w, h, layers;
	if (std::sscanf(line.c_str(), "%zu %zu %zu", &w, &h, &layers) != 3)
		throw std::invalid_argument("invalid map file");

	std::unique_ptr<Map> map(new Map(w, h, layers));

	std::getline(file, line);
	if (line.length())
		throw std::invalid_argument("invalid map file");

	size_t n = 1;
	std::unordered_map<char, size_t> tileset;

	for (std::getline(file, line); line.length(); std::getline(file, line)) {
		char c;
		unsigned rot;
		char *tile = nullptr;

		if (std::sscanf(line.c_str(), "%c %u %ms", &c, &rot, &tile) != 3 || rot >= 4) {
			std::free(tile);
			throw std::invalid_argument("invalid map file");
		}

		tileset.emplace(c, n++);
		map->tileset.emplace_back(tile, rot);
		std::free(tile);
	}

	for (size_t i = 0; i < h; i++) {
		std::getline(file, line);
		if (line.length() != w)
			throw std::invalid_argument("invalid map file");

		for (size_t j = 0; j < w; j++) {
			if (line[j] == '1')
				map->setCollisionAt(Position<int>{int(j), int(i)}, CollisionType::EMPTY);
			else if (line[j] != '0')
				throw std::invalid_argument("invalid map file");
		}
	}

	for (size_t layer = 0; layer < layers; layer++) {
		std::getline(file, line);
		if (line.length())
			throw std::invalid_argument("invalid map file");

		for (size_t i = 0; i < h; i++) {
			std::getline(file, line);
			if (line.length() != w)
				throw std::invalid_argument("invalid map file");

			for (size_t j = 0; j < w; j++) {
				auto it = tileset.find(line[j]);
				if (it != tileset.end())
					map->setTileAt(layer, Position<int>{int(j), int(i)}, it->second);
			}
		}
	}

	return map;
}

}

}
